// ========================================================================================= //
//                                                                                           //
//                                   TCP                                //
//                                     2017-03-27                                      //
//                                                                                           //
//                                       HiAsm                                         //
//                                     www.hiasm.com                                         //
//                                                                                           //
//        : Netspirit                                                                   //
//        E-mail: netspirit@meta.ua                                                          //
//                                                                                           //
//                                                       //
//                                                                                           //
//        1)    " ".                      //
//              .                                                //
//        2)          ,                     //
//                ,                  //
//            .                                                                     //
//        3)                             //
//           (  )   .                            //
//                                //
//                .                                //
//        4)                          //
//                                   //
//                  .                   //
//        5)        (, )         //
//                .                                //
//                                                                                           //
//                        ,                              //
//                                                                  //
//                                                                                           //
//                                           //
//                                                                                           //
// ========================================================================================= //


unit NSTCP;


interface


uses 
  Windows, Winsock, Kol, NSThreads;

type

  
  //      
  TReceiveBuffer = class
      FCurrentSize: Integer; //   FBuffer
      FBuffer: Pointer;
    public
      Len: Integer; //   
      destructor Destroy; override;
      function EnsureBuffer(SizeNeeded: Integer): Pointer; //      SizeNeeded
      function GetBuffer(var CurrLen: Integer): Pointer; //      
      function GetString: string; //   FBuffer     Len
  end;
  
  
  //        .
  //       .
  
  TSendData = class(TSharedObject)
    protected
      FSended: Integer;
      FLength: Integer;
      
    public
      Stopped: Boolean;
      KeepInQueue: Boolean; //       0

      //    ( ,      0). 0 -  
      function CurrLength: Integer; virtual;
      
      // ,    .    .
      procedure Finish; virtual;
      
      //   ,     ,   Buffer (  NeedLen).
      //  0 -   .  KeepInQueue=False        ().
      //  Stopped = True    0
      // ,   ,     .
      function GetNextData(var Buffer: Pointer; NeedLen: Integer): Integer; virtual;   
      
      //        
      procedure Stop; virtual;
      
      //     () 
      function Sended: Integer; virtual;
  end;
  
  //     TSendData
  TSendQueue = class
    private
      FCSList: TRTLCriticalSection;
      FDataToSend: PList;

      function GetItem(Index: Integer): TSendData;
      function GetHasActive: Boolean;
      function GetCount: Integer;
    public
      property HasActive: Boolean read GetHasActive; //     - 
      //    .      -   Lock
      property ItemCount: Integer read GetCount; 
      
      property Items[Index: Integer]: TSendData read GetItem; //   Lock

      constructor Create;
      destructor Destroy; override;
      
      //    .     /     
      procedure AddData(Data: TSendData);
      procedure Delete(Data: TSendData); overload; //   Lock
      procedure Delete(Index: Integer); overload; //   Lock
      
      //     //   
      procedure Lock; 
      
      //    .    -  nil.    -  .
      function NextData: TSendData;
      
      procedure StopAll; //       
      procedure UnLock; //      Lock,     
  end;
  
  TCustomConnection = class;
  
  //     
  TSendWorker = class(TCustomThread)
    private
      FConnList: PList;
      FCSList: TRTLCriticalSection;
      FTmpConn: TCustomConnection;
      FTmpData: TSendData;
      FEvSendReady: THandle;
      FNextIndex: Integer; //     
      
      procedure Execute; override;
      procedure _SyncDisconnect(Param: Pointer);
      procedure _SyncOnProgress(Param: Pointer);
      procedure _SyncOnSend(Param: Pointer);
      procedure _SyncSendError(Param: Pointer);
      function GetCount: Integer;
      function SelectNextConnection: TCustomConnection; // .      nil,  .   .
    public      
      property Count: Integer read GetCount;
      
      constructor Create;
      destructor Destroy; override;
      
      procedure AddConnection(Conn: TCustomConnection);
      procedure DeleteAll;
      procedure DeleteConnection(Conn: TCustomConnection);
      procedure StartSending;
      procedure StopSending;
      procedure StopAll;
  end;
  
  //   
  TSendManager = class
    private
      FWorkerList: PList;
      FCSWorkerLst: TRTLCriticalSection;
      procedure ClearWorkers;
      procedure StopSendAll;
    public
      MaxWorkers: Integer;

      constructor Create;
      destructor Destroy; override;

      procedure DeleteConnection(Conn: TCustomConnection);
      procedure SendAsync(Conn: TCustomConnection; Data: TSendData); //    Data (!)
      procedure StopSend(Conn: TCustomConnection);
  end;
  
  //      64 
  TConnectionListener = class(TCustomThread)
    private
      FConnections: PList;
      FCSList: TRTLCriticalSection;
      FEvRecvReady: THandle;
      FTmpConn: TCustomConnection;

      procedure Execute; override;
      procedure _SyncDisconnect(Param: Pointer);
      procedure _SyncReceive(Param: Pointer);

      function GetConnectionBySocket(Sock: TSocket): TCustomConnection;
      function GetCount: Integer;
      procedure GetSockets(var Sockets: TFDSet); //   Sockets      

    public
      property Count: Integer read GetCount; //     
      
      constructor Create;
      destructor Destroy; override;
      
      procedure AddConnection(Conn: TCustomConnection);
      procedure DeleteAll;
      procedure DeleteConnection(Conn: TCustomConnection);
      procedure StartReceiving;
      procedure StopReceiving;
  end;
  
  //      
  TReceivingManager = class
    private
      FListeners: PList;
      FCSListener: TRTLCriticalSection;
      
      procedure AdjustListeners;
      procedure ClearListeners;
      
    public
      constructor Create;
      destructor Destroy; override;
      
      procedure AddConnection(Conn: TCustomConnection);
      procedure DeleteConnection(Conn: TCustomConnection);
  end;
  
  // -   .
  TBasicConnection = class(TSharedObject)
    protected
      FSocket: TSocket;
      FRemoteAddress: string;
      FRemotePort: Integer;
      
      procedure Close; //virtual;
      function GetConnected: Boolean;
      function GetLocalAddress: string;
      function GetLocalPort: Integer;
      function GetSocketOptionInt(Option: Integer): Integer;
      procedure SetSocketOptionInt(Option, Value: Integer);
      
    public
      property Connected: Boolean read GetConnected;
      property LocalAddress: string read GetLocalAddress;
      property LocalPort: Integer read GetLocalPort;
      property RemoteAddress: string read FRemoteAddress;
      property RemotePort: Integer read FRemotePort;
      property SendTimeout: Integer index SO_SNDTIMEO read GetSocketOptionInt write SetSocketOptionInt; //      .         -    
      property Socket: TSocket read FSocket; //  
      
      destructor Destroy; override;
      
      //     
      //    ReadLen     
      //  RECV_SUCCESS  RECV_ERROR,  ReadLen -   
      function ReadData(Buffer: Pointer; var ReadLen: Integer): Integer;
      
      //    .  :
      // RECV_SUCCESS -  .  AvailLen   0 -    .
      // RECV_TIMEOUT -      . TimeOut -  .
      // RECV_ERROR -  , AvailLen   
      function QueryDataAvail(var AvailLen: Integer; TimeOut: Integer): Integer;
      
      //    .  0  ,   
      function WriteData(Buffer: Pointer; BufLen: Integer): Integer;
  end;
  
  
  //   TCP 
  TCustomConnection = class(TBasicConnection)
    private
      FCSDisconnect: TRTLCriticalSection;
      function GetSending: Boolean;
      
    protected
      FAbortSend: Boolean;
      FDestroying: Boolean;
      FSending: Boolean;
      FSendingAsync: Boolean;
      
      Listener: TConnectionListener;
      SendQueue: TSendQueue;
      SendWorker: TSendWorker;
      
      procedure AfterClose(Reason: Integer); virtual;
      procedure BeforeClose(Reason: Integer); virtual;
      procedure DataReceived(Data: TReceiveBuffer); virtual;
      procedure InternalClose(Reason: Integer);
      procedure SendResult(Data: TSendData; ResultCode: Integer); virtual; // ResultCode = 0 -  ,   
      
    public
      AsyncEvents: Boolean;

      OnProgress: procedure (Sender: TCustomConnection; Data: TSendData) of object; //   
      
      property IsSending: Boolean read GetSending;
      
      constructor Create;
      destructor Destroy; override;
      
      procedure Disconnect; //  
      function Send(Data: TSendData): Integer; //      Data      IncRef
      procedure SendAsync(Data: TSendData);
      procedure StartReceiving; //    
      procedure StopReceiving; //    
      procedure StopSend; //   
  end;

  //    TCP
  TCustomTCPClient = class(TCustomConnection)
    public
      function Connect(IP: string; Port: Integer): Integer; //     
  end;

  
  //    TCP
  TBasicTCPServer = class
    private
      FAcceptThread: TThread;
      FSocket: TSocket;
      FRemAddr, FRemPort: Integer;

      function _ExecuteAccept(T: TCustomThread): Integer;
      procedure _SyncClientConnect(Param: Pointer); // Param  TSocket
      function GetIsListening: Boolean;
      
    protected  
      procedure ConnectionAccepted(ClientSocket: TSocket; RemAddr: Integer; RemPort: Integer); virtual;
      
    public
      AsyncEvents: Boolean;
      
      property IsListening: Boolean read GetIsListening;
      
      destructor Destroy; override;
      
      function StartServer(Port: Integer): Integer; //     
      procedure StopListen; //  
  end;
  
  TCustomTCPServer = class;
  
  // ,    TCP 
  TCustomClientConnection = class(TCustomConnection)
    protected
      Server: TCustomTCPServer;
      procedure AfterClose(Reason: Integer); override;
    public
      constructor Create(Socket: TSocket; RemIP: Integer; RemPort: Integer = 0);
  end;
  
  //  TCP  
  TCustomTCPServer = class(TBasicTCPServer)
    private
      FConnList: PList;
      FCSList: TRTLCriticalSection;

      function GetCount: Integer;
      
    protected
      //      .       .
      //   ,     .
      procedure AfterNewConnection(Conn: TCustomClientConnection); virtual;
      
      //      .          TCustomClientConnection
      //  nil,   
      function BeforeNewConnection(ClientSocket: TSocket; RemAddr: Integer; RemPort: Integer): TCustomClientConnection; virtual;
      
      procedure ConnectionAccepted(ClientSocket: TSocket; RemAddr: Integer; RemPort: Integer); override;
      procedure ConnectionAdd(Conn: TCustomClientConnection);
      procedure ConnectionDelete(Conn: TCustomClientConnection); //   (!)
      
    public
      property Count: Integer read GetCount; //   

      constructor Create;
      destructor Destroy; override;

      procedure CloseAll; //    (!)
      function ConnectionExists(Conn: TCustomClientConnection): Boolean;
      procedure GetConnections(OutConn: PList = nil; OutIP: PStrList = nil); //   ,    
      //   ,   -      ().
      function HoldConnection(Conn: TCustomClientConnection): TCustomClientConnection;
      procedure StopSendAll;
      procedure StopServer; //      
  end;
  

  // =============   ============= //
  
  //       
  TSharedString = class(TSharedObject)
    private
      FString: string;
      
    public
      property GetString: string read FString;
      
      constructor Create(const S: string);
      
      function GetBuffer(Offset: Integer): Pointer; //     FString[1+Offset]
      function GetLength: Integer;
  end;
  
  //  TSendData   
  // TODO:      TSharedString
  TSendDataString = class(TSendData)
    private
      FData: TSharedString;
    public
      constructor Create(Data: TSharedString);
      destructor Destroy; override;
      procedure Finish; override;
      function GetNextData(var Buffer: Pointer; NeedLen: Integer): Integer; override;
  end;
  
  //  TCP 
  TTCPClient = class(TCustomTCPClient)
    protected
      procedure AfterClose(Reason: Integer); override;
      procedure DataReceived(Data: TReceiveBuffer); override;
      procedure SendResult(Data: TSendData; ResultCode: Integer); override;
    public
      OnDisconnect: procedure (Sender: TCustomConnection) of object;
      OnReceive: procedure (Sender: TCustomConnection; Data: string) of object;
      OnErrorSend: procedure (Sender: TCustomConnection; ErrorCode: Integer) of object;
      OnSend: procedure (Sender: TCustomConnection; Data: TSendData) of object;
  end;
  
  TClientConnection = class(TCustomClientConnection)
    protected
      procedure AfterClose(Reason: Integer); override;
      procedure SendResult(Data: TSendData; ResultCode: Integer); override;
    protected
      procedure DataReceived(Data: TReceiveBuffer); override;
  end;
  
  //  TCP 
  TTCPServer = class(TCustomTCPServer)
    private
      function BeforeNewConnection(ClientSocket: TSocket; RemAddr: Integer; RemPort: Integer): TCustomClientConnection; override;
      procedure AfterNewConnection(Conn: TCustomClientConnection); override;
    public
      OnClientConnect: procedure (Conn: TCustomClientConnection) of object;
      OnClientDisconnect: procedure (Conn: TCustomClientConnection) of object;
      OnErrorSend: procedure (Conn: TCustomClientConnection; ErrorCode: Integer) of object;
      OnProgress: procedure (Conn: TCustomConnection; Data: TSendData) of object;
      OnReceive: procedure (Conn: TCustomClientConnection; S: string) of object;
      OnSend: procedure (Conn: TCustomClientConnection; Data: TSendData) of object;
      
      procedure SendAll(Data: TSharedString);
      procedure SendAllAsync(Data: TSharedString; Queue: Boolean = True);
  end;
  
  
  


  //   WinSock.TFDSet   
  TSocketSet = record
    Count: Integer; //  1
    Socket: TSocket;
  end;
  
  function IPv4AddressToString(Addr: Integer): string;
  

  
const
  SD_RECEIVE     = 0;
  SD_SEND        = 1;
  SD_BOTH        = 2;
  
  RECV_ERROR   = -1;
  RECV_SUCCESS =  0;
  RECV_TIMEOUT =  1;

  DISCONNECT_REASON_MANUAL     = 0;
  DISCONNECT_REASON_NETWORK    = 1;
  DISCONNECT_REASON_DESTROYING = 2;


implementation




const
  //     .     .
  //    ()     .
  //   ,        GetSocketOptionInt(SO_RCVBUF) ( 8 KB)
  SEND_FRAGMENT_SIZE = 16*1024;
  
  //        .   FD_SETSIZE (=64).
  //          
  MAX_SOCKETS_PER_THREAD = FD_SETSIZE;
  
  //       (      )
  MAX_SENDING_THREADS = 5;
  
  //      
  SEND_WAITING_TIME = 10000;
  
  //        
  RECV_WAITING_TIME = 10000;

var 
  wsaData: TWSAData;

  SendManager: TSendManager;
  ReceivingManager: TReceivingManager;




function IPv4AddressToString(Addr: Integer): string;
var
  IPRecord: record
    c1, c2, c3, c4: byte; 
  end absolute Addr;
begin
  with IPRecord do
    Result := Int2Str(c1) + '.' + Int2Str(c2) + '.' + Int2Str(c3) + '.' + Int2Str(c4);
end;



// ====== TReceiveBuffer ====== //

destructor TReceiveBuffer.Destroy;
begin
  if FCurrentSize > 0 then FreeMem(FBuffer, FCurrentSize);
  FCurrentSize := 0;
  inherited;
end;

function TReceiveBuffer.EnsureBuffer(SizeNeeded: Integer): Pointer;
begin
  if FCurrentSize < SizeNeeded then
  begin
    if FCurrentSize > 0 then FreeMem(FBuffer, FCurrentSize);
    GetMem(FBuffer, SizeNeeded);
    FCurrentSize := SizeNeeded;
  end;
  Result := FBuffer;
end;

function TReceiveBuffer.GetBuffer(var CurrLen: Integer): Pointer;
begin
  Result := FBuffer;
  CurrLen := Len;
end;

function TReceiveBuffer.GetString: string;
begin
  Result := '';
  if (FCurrentSize > 0) and (Len > 0) then // TODO:  Len <= FCurrentSize
  begin
    SetLength(Result, Len);
    Move(FBuffer^, Result[1], Len);
  end;
end;


// ====== TSendData ====== //

function TSendData.CurrLength: Integer;
begin
  Result := FLength;
end;

procedure TSendData.Finish;
begin

end;

function TSendData.GetNextData(var Buffer: Pointer; NeedLen: Integer): Integer;
begin
  Result := 0;
end;

function TSendData.Sended: Integer;
begin
  Result := FSended;
end;

procedure TSendData.Stop;
begin
  FLength := 0;
  Stopped := True;
end;

// ====== TSendQueue ====== //

constructor TSendQueue.Create;
begin
  inherited;
  InitializeCriticalSection(FCSList);
  FDataToSend := NewList;
end;

destructor TSendQueue.Destroy;
begin
  EnterCriticalSection(FCSList);
    StopAll;
    FDataToSend.Free;
  LeaveCriticalSection(FCSList);
  DeleteCriticalSection(FCSList);
  inherited;
end;

function TSendQueue.GetItem(Index: Integer): TSendData;
begin
  Result := TSendData(FDataToSend.Items[Index]);
end;

function TSendQueue.GetHasActive: Boolean;
var
  I: Integer;
begin
  Result := False;
  EnterCriticalSection(FCSList);
    for I := 0 to FDataToSend.Count - 1 do
    begin
      if TSendData(FDataToSend.Items[I]).CurrLength <> 0 then
      begin
        Result := True;
        Break;
      end;
    end;  
  LeaveCriticalSection(FCSList);
end;

function TSendQueue.GetCount: Integer;
begin
  Result := FDataToSend.Count;
end;

procedure TSendQueue.AddData(Data: TSendData);
begin
  if Data.CurrLength = 0 then
  begin
    Data.DecRef;
    Exit;
  end;
  
  EnterCriticalSection(FCSList);
    FDataToSend.Add(Data);
  LeaveCriticalSection(FCSList);
end;

procedure TSendQueue.Delete(Data: TSendData);
begin
  FDataToSend.Remove(Data);
end;

procedure TSendQueue.Delete(Index: Integer);
begin
  if (Index >= 0) and (Index < FDataToSend.Count) then FDataToSend.Delete(Index);
end;

procedure TSendQueue.Lock;
begin
  EnterCriticalSection(FCSList);
end;

function TSendQueue.NextData: TSendData;
var
  I: Integer;
begin
  Result := nil;
  EnterCriticalSection(FCSList);
    I := 0;
    while I < FDataToSend.Count do
    begin
      Result := FDataToSend.Items[I];
      if (Result.CurrLength <> 0) and (not Result.Stopped) then
      begin
        Result.IncRef;
        Break;
      end;
      
      if not Result.KeepInQueue then
      begin
        FDataToSend.Delete(I);
        Result.DecRef;
      end
      else
        Inc(I);
        
      Result := nil;
    end;
  LeaveCriticalSection(FCSList);
end;

procedure TSendQueue.StopAll;
var
  I: Integer;
  D: TSendData;
begin
  EnterCriticalSection(FCSList);
    for I := 0 to FDataToSend.Count - 1 do
    begin
      D := TSendData(FDataToSend.Items[I]);
      D.Stop;
      D.DecRef;
    end;
    FDataToSend.Clear;
  LeaveCriticalSection(FCSList);
end;

procedure TSendQueue.UnLock;
begin
  LeaveCriticalSection(FCSList);
end;

// ====== TSendWorker ====== //

constructor TSendWorker.Create;
begin
  inherited;
  InitializeCriticalSection(FCSList);
  FConnList := NewList;
  FConnList.AddBy := 20;
  FEvSendReady := CreateEvent(nil, False{ManualReset}, False, nil);
end;

destructor TSendWorker.Destroy;
begin
  EnterCriticalSection(FCSList);
    DeleteAll;
    FConnList.Free;
  LeaveCriticalSection(FCSList);
  DeleteCriticalSection(FCSList);
  CloseHandle(FEvSendReady);
  inherited;
end;

procedure TSendWorker.Execute;
var 
  C: TCustomConnection;
  Dt: TSendData;
  SendLen: Integer;
  Buf: Pointer;
  Err: Integer;
begin
  //ResetEvent(FEvSendReady);
  
  while not Stopped do
  begin
    C := SelectNextConnection; //  
    
    if Stopped then
    begin
      if C <> nil then C.DecRef;
      Exit;
    end;
    
    //   ,         ,
    //     ,        
    if C = nil then
    begin    
      Err := WaitForSingleObject(FEvSendReady, SEND_WAITING_TIME);

      if Stopped then Exit;
      
      if Err = WAIT_TIMEOUT then Break
      else
        //ResetEvent(FEvSendReady);
        Continue;
    end;
    // ---------------
    
    Dt := C.SendQueue.NextData; //   Dt   TSendQueue

    if Dt = nil then //     SelectNextConnection   
    begin
      C.FSendingAsync := False;
      C.DecRef; // "" 
      Continue;
    end;

    SendLen := Dt.GetNextData(Buf, SEND_FRAGMENT_SIZE);
    
    if SendLen = 0 then //     Dt := C.SendQueue.NextData   
    begin
      Dt.DecRef; // "" 
      C.DecRef;
      Continue;
    end;


    //  .    ...
    Err := C.WriteData(Buf, SendLen);

    if Stopped then
    begin
      Dt.DecRef;
      C.DecRef;
      Break;
    end;

    if C.FAbortSend then
    begin
      Dt.DecRef;
      C.DecRef;
      Continue;
    end;
    
    FTmpConn := C;
    

    
    // -----    
    if Err <> 0 then
    begin
      FTmpData := Dt;
      C.SendQueue.StopAll; //     (   ?)
      C.FSendingAsync := False;
      
      //     Err = WSAESHUTDOWN
      //     
      if C.Connected then
      begin
        if C.AsyncEvents then
          _SyncSendError(Pointer(Err))
        else
          Synchronize(_SyncSendError, Pointer(Err));
      end;
      
      Dt.DecRef;
      
      if Stopped then
      begin
        C.DecRef;
        Break;
      end;
      
      if C.FAbortSend then
      begin
        C.DecRef;
        Continue;
      end;
        
      if (Err = WSAECONNRESET) or (Err = WSAECONNABORTED) then
      begin
        if C.AsyncEvents then
          _SyncDisconnect(C)
        else
          Synchronize(_SyncDisconnect, C);
      end;
      
      C.DecRef;
      Continue;
    end;
    // ---------------
    
    //if not C.SendQueue.HasActive then C.FSendingAsync := False;

    // -----  
    if Assigned(C.OnProgress) then
    begin
      if C.AsyncEvents then
        _SyncOnProgress(Dt)
      else
        Synchronize(_SyncOnProgress, Dt);
      
      if Stopped then
      begin
        Dt.DecRef;
        C.DecRef;
        Break;
      end;
    end;
    // ---------------

 
    if C.FAbortSend then //    -     .    OnSend  ?
    begin
      Dt.DecRef;
      C.DecRef;
      Continue;
    end;

    if not C.SendQueue.HasActive then C.FSendingAsync := False;

    // -----  
    if (Dt.CurrLength = 0) {and (not Dt.Stopped)} then 
    begin
      Dt.Finish;
      if C.AsyncEvents then
        _SyncOnSend(Dt)
      else
        Synchronize(_SyncOnSend, Dt);
    end;

    Dt.DecRef;
    C.DecRef;
    // ------------
    
  end;
  
end;

procedure TSendWorker._SyncDisconnect(Param: Pointer);
begin
  //DeleteConnection(TCustomConnection(Param)); //  InternalClose
  TCustomConnection(Param).InternalClose(DISCONNECT_REASON_NETWORK);
end;

procedure TSendWorker._SyncOnProgress(Param: Pointer);
begin
  FTmpConn.OnProgress(FTmpConn, TSendData(Param));
end;

procedure TSendWorker._SyncOnSend(Param: Pointer);
begin
  FTmpConn.SendResult(TSendData(Param), 0);
end;

procedure TSendWorker._SyncSendError(Param: Pointer);
begin
  FTmpConn.SendResult(FTmpData, Integer(Param));
end;

function TSendWorker.GetCount: Integer;
begin
  //EnterCriticalSection(FCSList);
    Result := FConnList.Count;
  //LeaveCriticalSection(FCSList);
end;

procedure TSendWorker.AddConnection(Conn: TCustomConnection);
begin
  EnterCriticalSection(FCSList);
    if (Conn.SendWorker = nil) {and (FConnList.IndexOf(Conn) = -1)} then
    begin
      //Conn.IncRef; <---  SendManager
      Conn.SendWorker := Self;
      FConnList.Add(Conn);
      if Conn.SendQueue.HasActive then StartSending;
    end;
  LeaveCriticalSection(FCSList);
end;

procedure TSendWorker.DeleteConnection(Conn: TCustomConnection);
var
  Idx: Integer;
begin
  EnterCriticalSection(FCSList);
    Idx := FConnList.IndexOf(Conn);
    if Idx > -1 then
    begin
      Conn.SendQueue.StopAll;
      Conn.FSendingAsync := False;
      Conn.SendWorker := nil;
      FConnList.Delete(Idx);
      Conn.DecRef;
      //if FConnList.Count = 0 then StopSending;
    end;
  LeaveCriticalSection(FCSList);
end;

procedure TSendWorker.DeleteAll;
var
  I: Integer;
  C: TCustomConnection;
begin
  StopAll;
  EnterCriticalSection(FCSList);
    for I := 0 to FConnList.Count - 1 do
    begin
      C := FConnList.Items[I];
      C.FSendingAsync := False;
      C.SendWorker := nil;
      C.DecRef;
    end;
    FConnList.Clear;
  LeaveCriticalSection(FCSList);
end;

function TSendWorker.SelectNextConnection: TCustomConnection;
var
  I, C: Integer;
begin
  I := 0;
  EnterCriticalSection(FCSList);
    C := FConnList.Count;
    while I < C do
    begin
      if FNextIndex >= C then FNextIndex := 0;
      Result := TCustomConnection(FConnList.Items[FNextIndex]);
      Inc(FNextIndex);
      if Result.SendQueue.HasActive and Result.Connected then
      begin
        Result.IncRef;
        LeaveCriticalSection(FCSList);
        Exit;
      end;
      Inc(I);
    end;
  LeaveCriticalSection(FCSList);
  Result := nil;
end;

procedure TSendWorker.StartSending;
begin
  EnterCriticalSection(FCSList);
    if Terminated or Suspended then Restart
    else  
      // <===     ,     
      //      ...
      SetEvent(FEvSendReady); //   ,   
  LeaveCriticalSection(FCSList);
end;

procedure TSendWorker.StopAll;
var
  I: Integer;
  C: TCustomConnection;
begin
  StopSending;
  EnterCriticalSection(FCSList);
    for I := 0 to FConnList.Count - 1 do
    begin
      C := FConnList.Items[I];
      C.SendQueue.StopAll;
    end;
  LeaveCriticalSection(FCSList);
end;

procedure TSendWorker.StopSending;
begin
  Stop;
  SetEvent(FEvSendReady);
end;  


// ====== TSendManager ====== //

constructor TSendManager.Create;
begin
  inherited;
  InitializeCriticalSection(FCSWorkerLst);
  FWorkerList := NewList;
  MaxWorkers := MAX_SENDING_THREADS;
end;

destructor TSendManager.Destroy;
begin
  EnterCriticalSection(FCSWorkerLst);
    StopSendAll;
    ClearWorkers;
    FWorkerList.Free;
  LeaveCriticalSection(FCSWorkerLst);
  DeleteCriticalSection(FCSWorkerLst);
  inherited;
end;

procedure TSendManager.ClearWorkers;
var
  I: Integer;
begin
  for I := 0 to FWorkerList.Count - 1 do
  begin
    with TSendWorker(FWorkerList.Items[I]) do
    begin
      StopSending; //   
      DecRef; //  . TSendWorker    ,     ( )
    end;
  end;
  FWorkerList.Clear;
end;

procedure TSendManager.DeleteConnection(Conn: TCustomConnection);
begin
  if Conn.SendWorker <> nil then Conn.SendWorker.DeleteConnection(Conn);
end;

procedure TSendManager.SendAsync(Conn: TCustomConnection; Data: TSendData);
var
  I, MinC: Integer;
  W, OptW: TSendWorker;
begin

  if Data.CurrLength = 0 then
  begin
    Data.DecRef;
    Exit;
  end;

  Conn.FSendingAsync := True;
  Conn.SendQueue.AddData(Data);
  
  if Conn.SendWorker <> nil then //     
  begin
    Conn.SendWorker.StartSending;
    Exit;
  end;

  Conn.IncRef;
  
  //      
  EnterCriticalSection(FCSWorkerLst);
  
    if FWorkerList.Count < MaxWorkers then
    begin
      W := TSendWorker.Create;
      FWorkerList.Add(W);
      W.AddConnection(Conn);
    end
    else
    begin
      //    Worker,   
      MinC := 99999;
      OptW := nil;
      for I := 0 to FWorkerList.Count - 1 do
      begin
        W := TSendWorker(FWorkerList.Items[I]);
        if W.Count < MinC then
        begin
          MinC := W.Count;
          OptW := W;
        end;
      end;
      if OptW <> nil then // ,     ,  -  ...
      begin
        OptW.AddConnection(Conn);
      end;
    end;
    
  LeaveCriticalSection(FCSWorkerLst);
end;

procedure TSendManager.StopSend(Conn: TCustomConnection);
begin
  Conn.SendQueue.StopAll;
end;

procedure TSendManager.StopSendAll;
var
  I: Integer;
  W: TSendWorker;
begin
  EnterCriticalSection(FCSWorkerLst);
    for I := 0 to FWorkerList.Count - 1 do
    begin
      W := TSendWorker(FWorkerList.Items[I]);
      W.StopAll;
      W.StopSending;
    end;
  LeaveCriticalSection(FCSWorkerLst);
end;


// ====== TConnectionListener ====== //
 
constructor TConnectionListener.Create;
begin
  inherited;
  FConnections := NewList;
  FConnections.Capacity := MAX_SOCKETS_PER_THREAD; //   ,     
  InitializeCriticalSection(FCSList);
  FEvRecvReady := CreateEvent(nil, False{ManualReset}, False, nil);
end;

destructor TConnectionListener.Destroy;
begin
  StopReceiving;
  DeleteAll;
  EnterCriticalSection(FCSList);
    FConnections.Free;
  LeaveCriticalSection(FCSList);
  DeleteCriticalSection(FCSList);
  CloseHandle(FEvRecvReady);
  inherited;
end;


procedure TConnectionListener.Execute;
var
  Buf: Pointer;
  ByteReceived, Avail, I: Integer;
  Res: TReceiveBuffer;
  TimeVal: TTimeVal;
  FDRead: TFDSet;
  
  Sock: TSocket;
begin
  //ResetEvent(FEvRecvReady);
  
  Res := TReceiveBuffer.Create; //       
  
  TimeVal.tv_sec := 2; //    - 2 
  TimeVal.tv_usec := 0;

  while not Stopped do
  begin

    //     
    GetSockets(FDRead);
    
    //   ,     ,
    //        
    if FDRead.fd_count = 0 then
    begin
      I := WaitForSingleObject(FEvRecvReady, RECV_WAITING_TIME);
      if Stopped then Break;
      
      if I = WAIT_TIMEOUT then Break
      else
        //ResetEvent(FEvRecvReady);
        Continue;
    end;
    // ---------------
    
    //    (   )

    Avail := select(0, @FDRead, nil, nil, @TimeVal);
    case Avail of
      SOCKET_ERROR:
        begin
          // - 
          Break;
        end;
      0: Continue; // 
    end;
    
    for I := 0 to FDRead.fd_count-1 do
    begin
      if Stopped then Break;
      Sock := FDRead.fd_array[I];
      
      ioctlsocket(Sock, FIONREAD, Avail);
      
      //     
      if Avail < 1 then //  ,   ,   0 -  
      begin
        if not Stopped then
        begin
          FTmpConn := GetConnectionBySocket(Sock); //   
          if FTmpConn <> nil then  // ,   
          begin
            if FTmpConn.AsyncEvents then
              _SyncDisconnect(FTmpConn)
            else
              Synchronize(_SyncDisconnect, FTmpConn);
            FTmpConn.DecRef;
          end;
        end
        else Break;
        Continue;
      end;

      //  .          
      Buf := Res.EnsureBuffer(Avail);
      
      //  .     select,  recv   
      //            
      ByteReceived := recv(Sock, Buf^, Avail, 0);
      if ByteReceived < 1 then // 0 -   , -1 -  
      begin 
        if not Stopped then
        begin
          FTmpConn := GetConnectionBySocket(Sock);
          if FTmpConn <> nil then  // ,   
          begin
            if FTmpConn.AsyncEvents then
              _SyncDisconnect(FTmpConn)
            else
              Synchronize(_SyncDisconnect, FTmpConn);
            FTmpConn.DecRef;
          end;
        end;
        Continue;
      end;

      if not Stopped then
      begin
        FTmpConn := GetConnectionBySocket(Sock);
        if FTmpConn <> nil then  // ,   
        begin
          Res.Len := ByteReceived; //  

          //     
          if FTmpConn.AsyncEvents then
            _SyncReceive(Res)
          else
            Synchronize(_SyncReceive, Res);
          
          FTmpConn.DecRef;
        end;
      end;

    end; // for
  
  end; // while
    
  Res.Free;
end;

procedure TConnectionListener._SyncDisconnect(Param: Pointer);
begin
  TCustomConnection(Param).InternalClose(DISCONNECT_REASON_NETWORK);
end;

procedure TConnectionListener._SyncReceive(Param: Pointer);
begin
  FTmpConn.DataReceived(TReceiveBuffer(Param));
end;

function TConnectionListener.GetConnectionBySocket(Sock: TSocket): TCustomConnection;
var
  I: Integer;
label
  finish;
begin
  EnterCriticalSection(FCSList);
    for I := 0 to FConnections.Count - 1 do
    begin
      Result := TCustomConnection(FConnections.Items[I]);
      if Result.Socket = Sock then
      begin
        Result.IncRef;
        goto finish;
      end;
    end;
    Result := nil;
finish:
  LeaveCriticalSection(FCSList);
end;

function TConnectionListener.GetCount: Integer;
begin
  Result := FConnections.Count;  
end;

procedure TConnectionListener.GetSockets(var Sockets: TFDSet);
var
  I: Integer;
  C: TCustomConnection;
begin
  EnterCriticalSection(FCSList);
    Sockets.fd_count := 0;
    for I := 0 to FConnections.Count - 1 do
    begin
      C := TCustomConnection(FConnections.Items[I]);
      if C.Connected then
      begin
        Sockets.fd_array[Sockets.fd_count] := C.Socket;
        Inc(Sockets.fd_count);
      end;
    end;
  LeaveCriticalSection(FCSList);
end;

procedure TConnectionListener.AddConnection(Conn: TCustomConnection);
begin
  EnterCriticalSection(FCSList);
    //if (FConnections.Count < MAX_SOCKETS_PER_THREAD) and (FConnections.IndexOf(Conn) = -1) then
    //begin
      //Conn.IncRef; <===  TReceivingManager
      FConnections.Add(Conn);
      Conn.Listener := Self;
      StartReceiving;
    //end;
  LeaveCriticalSection(FCSList);
end;

procedure TConnectionListener.DeleteAll;
var
  I: Integer;
  C: TCustomConnection;
begin
  EnterCriticalSection(FCSList);
    for I := 0 to FConnections.Count - 1 do
    begin
      C := TCustomConnection(FConnections.Items[I]);
      C.DecRef;
    end;
    FConnections.Clear;
  LeaveCriticalSection(FCSList);
end;

procedure TConnectionListener.DeleteConnection(Conn: TCustomConnection);
begin
  EnterCriticalSection(FCSList);
    //if Conn.Listener = Self then
    //begin
      Conn.Listener := nil;
      FConnections.Remove(Conn);
      Conn.DecRef;
      //if FConnections.Count = 0 then StopReceiving; //    
    //end;
  LeaveCriticalSection(FCSList);
end;

procedure TConnectionListener.StartReceiving;
begin
  if FConnections.Count = 0 then Exit;
  
  if Terminated or Suspended then Restart
  else
    SetEvent(FEvRecvReady);
end;

procedure TConnectionListener.StopReceiving;
begin
  Stop;
  SetEvent(FEvRecvReady); //      
end;


// ====== TReceivingManager ====== //

constructor TReceivingManager.Create;
begin
  inherited;
  FListeners := NewList;
  InitializeCriticalSection(FCSListener);
end;

destructor TReceivingManager.Destroy;
begin
  ClearListeners;
  FListeners.Free;
  DeleteCriticalSection(FCSListener);
  inherited;
end;

procedure TReceivingManager.AddConnection(Conn: TCustomConnection);
var
  I: Integer;
  L: TConnectionListener;
label
  finish;
begin
  if Conn.Listener <> nil then Exit;

  Conn.IncRef;
  
  EnterCriticalSection(FCSListener);
    //  ,    
    for I := 0 to FListeners.Count - 1 do
    begin
      L := FListeners.Items[I];
      if L.Count < MAX_SOCKETS_PER_THREAD then
      begin
        L.AddConnection(Conn);
        goto finish;
      end;
    end;
    
    //      - 
    L := TConnectionListener.Create;
    FListeners.Add(L);
    L.AddConnection(Conn);
    
finish:
  LeaveCriticalSection(FCSListener);
end;

procedure TReceivingManager.AdjustListeners;
var
  I, K: Integer;
  L: TConnectionListener;
begin
  EnterCriticalSection(FCSListener);
    K := 0;
    I := 0;
    while I < FListeners.Count do
    begin
      L := FListeners.Items[I];
      if L.Count = 0 then
      begin
        if K = 1 then //    Listener   
        begin
          FListeners.Delete(I);
          L.StopReceiving; //  
          L.DecRef; //  . TConnectionListener    ,     ( )
          Continue;
        end
        else
          Inc(K);
      end;
      Inc(I);
    end;
  LeaveCriticalSection(FCSListener);
end;

procedure TReceivingManager.ClearListeners;
var
  I: Integer;
begin
  EnterCriticalSection(FCSListener);
    for I := 0 to FListeners.Count - 1 do
    begin
      with TConnectionListener(FListeners.Items[I]) do
      begin
        StopReceiving; //  
        DecRef; //  . TConnectionListener    ,     ( )
      end;
    end;
    FListeners.Clear;
  LeaveCriticalSection(FCSListener);
end;

procedure TReceivingManager.DeleteConnection(Conn: TCustomConnection);
var
  L: TConnectionListener;
begin
  L := Conn.Listener;
  if L <> nil then
  begin
    L.DeleteConnection(Conn);
    if L.Count = 0 then AdjustListeners;
  end;
end;

  
// ====== TBasicConnection ====== //

destructor TBasicConnection.Destroy;
begin
  Close;
  inherited;
end;

procedure TBasicConnection.Close;
var
  S: TSocket;
begin
  if FSocket <> 0 then
  begin
    S := FSocket;
    FSocket := 0;
    shutdown(S, SD_BOTH);
    closesocket(S);
    
    FRemoteAddress := '';
    FRemotePort := 0;
  end;
end;

function TBasicConnection.GetConnected: Boolean;
begin
  Result := FSocket <> 0;
end;

function TBasicConnection.GetLocalAddress: string;
var
  nm: TSockAddr;
  len: Integer;
begin
  len := SizeOf(TSockAddr);
  if getsockname(FSocket, nm, len) = 0 then
    Result := IPv4AddressToString(Integer(nm.sin_addr))
  else
    Result := '';
end;

function TBasicConnection.GetLocalPort: Integer;
var
  nm: TSockAddr;
  len: Integer;
begin
  len := SizeOf(TSockAddr);
  if getsockname(FSocket, nm, len) = 0 then
    Result := ntohs(nm.sin_port)
  else
    Result := 0;
end;

function TBasicConnection.GetSocketOptionInt(Option: Integer): Integer;
var
  Sz: Integer;
begin
  Result := 0;
  Sz := SizeOf(Integer);
  getsockopt(FSocket, SOL_SOCKET, Option, @Result, Sz);
end;

procedure TBasicConnection.SetSocketOptionInt(Option, Value: Integer);
begin
  setsockopt(FSocket, SOL_SOCKET, Option, PChar(@Value), SizeOf(Integer));
end;

function TBasicConnection.ReadData(Buffer: Pointer; var ReadLen: Integer): Integer;
var
  ByteReceived: Integer;
begin
  ByteReceived := recv(FSocket, Buffer^, ReadLen, 0);
  if ByteReceived = SOCKET_ERROR then
  begin
    Result := RECV_ERROR;
    ReadLen := WSAGetLastError;
  end
  else
  begin
    Result := RECV_SUCCESS;
    ReadLen := ByteReceived;
  end;
end;

function TBasicConnection.QueryDataAvail(var AvailLen: Integer; TimeOut: Integer): Integer;
var
  FDRead: TSocketSet;
  TimeVal: TTimeVal;
begin
  AvailLen := 0;
  
  TimeVal.tv_sec := TimeOut div 1000; // 
  TimeVal.tv_usec := (TimeOut mod 1000) * 1000; // 

  FDRead.Count := 1;
  FDRead.Socket := FSocket;
  
  case select(0, PFDSet(@FDRead), nil, nil, @TimeVal) of
    SOCKET_ERROR:
      begin
        Result := RECV_ERROR;
        AvailLen := WSAGetLastError;
      end;
    0:
      Result := RECV_TIMEOUT;
    else
    begin
      Result := RECV_SUCCESS;
      ioctlsocket(FDRead.Socket, FIONREAD, AvailLen);
    end;
  end;
    
end;

function TBasicConnection.WriteData(Buffer: Pointer; BufLen: Integer): Integer;
begin
  Result := 0;
  
  if FSocket = 0 then Exit;

  if Winsock.send(FSocket, Buffer^, BufLen, 0) = SOCKET_ERROR then
    Result := WSAGetLastError;
end;


// ====== TCustomConnection ====== //

constructor TCustomConnection.Create;
begin
  inherited;
  //AsyncEvents := False;
  SendQueue := TSendQueue.Create;
  InitializeCriticalSection(FCSDisconnect);
end;

destructor TCustomConnection.Destroy;
begin
  if FDestroying then Exit; //       InternalClose->AfterClose  .
  FDestroying := True;

  InternalClose(DISCONNECT_REASON_DESTROYING);
  SendQueue.Free;
  DeleteCriticalSection(FCSDisconnect);
  inherited;
end;

procedure TCustomConnection.AfterClose(Reason: Integer);
begin

end;

procedure TCustomConnection.BeforeClose(Reason: Integer);
begin
  StopReceiving;
  StopSend;
end;

procedure TCustomConnection.DataReceived(Data: TReceiveBuffer);
begin

end;

procedure TCustomConnection.Disconnect;
begin
  InternalClose(DISCONNECT_REASON_MANUAL);
end;

function TCustomConnection.GetSending: Boolean;
begin
  Result := FSending or FSendingAsync;
end;

procedure TCustomConnection.InternalClose(Reason: Integer);
begin
  EnterCriticalSection(FCSDisconnect);
  if FSocket <> 0 then
  begin
    BeforeClose(Reason);
    Close;
    LeaveCriticalSection(FCSDisconnect);
    AfterClose(Reason);
  end
  else
    LeaveCriticalSection(FCSDisconnect);
end;

procedure TCustomConnection.SendResult(Data: TSendData; ResultCode: Integer);
begin

end;

function TCustomConnection.Send(Data: TSendData): Integer;
var
  Buf: Pointer;
  SendLen: Integer;
  Err: Integer absolute Result;
begin
  Result := 0;
  
  // IncRef     IsSending, FSocket
  if IsSending or (Data.CurrLength = 0) or (FSocket = 0) then
  begin
    Data.DecRef;
    Exit;
  end;
  
  IncRef; //         
  
  FSending := True;
  FAbortSend := False;
  
  while True do
  begin
    if FAbortSend then Break;
    
    SendLen := Data.GetNextData(Buf, SEND_FRAGMENT_SIZE);
    
    if SendLen = 0 then
    begin
      //FSending := False; //  Send  OnSend   ,    
      Data.Finish;
      SendResult(Data, 0);
      Break;
    end;
    
    //  .    ...
    Err := WriteData(Buf, SendLen);
    
    if FAbortSend then Break;
    
    //    
    if Err <> 0 then
    begin
      if Connected then SendResult(Data, Err);

      if (Err = WSAECONNRESET) or (Err = WSAECONNABORTED) then
        InternalClose(DISCONNECT_REASON_NETWORK);

      Break;
    end;

    if Assigned(OnProgress) then  OnProgress(Self, Data);      

  end;
  
  Data.DecRef;
  FSending := False;
  
  DecRef;
end;

procedure TCustomConnection.SendAsync(Data: TSendData);
begin
  if (FSocket = 0) or (FSending) then
  begin
    Data.DecRef;
    Exit;
  end;

  FAbortSend := False;
  SendManager.SendAsync(Self, Data);
end;

procedure TCustomConnection.StartReceiving;
begin
  if FSocket = 0 then Exit;
  ReceivingManager.AddConnection(Self);
end;

procedure TCustomConnection.StopReceiving;
begin
  ReceivingManager.DeleteConnection(Self);
end;

procedure TCustomConnection.StopSend;
begin
  FAbortSend := True;
  SendManager.DeleteConnection(Self);
end;

  
// ====== TCustomTCPClient ====== //

function TCustomTCPClient.Connect(IP: string; Port: Integer): Integer;
var 
  Addr: TSockAddr;
begin
  Result := 0;
  if FSocket <> 0 then Exit;
  
  FRemoteAddress := '';
  FRemotePort := 0;

  FSocket := Winsock.socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
  if FSocket = 0 then
  begin
    Result := WSAGetLastError;
    Exit;
  end;
  
  FillChar(Addr, SizeOf(TSockAddr), #0);
  Addr.sin_family := AF_INET;
  Addr.sin_port := htons(Port);
  Addr.sin_addr.S_addr := inet_addr(PChar(IP));
  
  if Winsock.connect(FSocket, Addr, SizeOf(Addr)) <> 0 then
  begin
    Result := WSAGetLastError;
    closesocket(FSocket);
    FSocket := 0;
  end
  else
  begin
    FRemoteAddress := IP;
    FRemotePort := Port;
  end;
end;


// ====== TBasicTCPServer ====== //

destructor TBasicTCPServer.Destroy;
begin
  StopListen;
  inherited;
end;

function TBasicTCPServer._ExecuteAccept(T: TCustomThread): Integer;
var
  ClientSocket: TSocket;
  Addr: TSockAddr;
  AddrLen: Integer;
  FDRead: TSocketSet;
  TimeVal: TTimeVal;
begin
  Result := 0;

  TimeVal.tv_sec := 2; //    - 2 
  TimeVal.tv_usec := 0;
  
  while (not T.Stopped) and (FSocket <> 0) do
  begin
    
    FDRead.Count := 1;
    FDRead.Socket := FSocket;
    
    //    (   )

    case select(0, PFDSet(@FDRead), nil, nil, @TimeVal) of
      SOCKET_ERROR:
        begin
          if not T.Stopped then StopListen;
          Break;
        end;
      0: Continue; // 
    end;

    //   - 
    
    FillChar(Addr, SizeOf(TSockAddr), #0);
    AddrLen := SizeOf(TSockAddr);
    
    ClientSocket := accept(FSocket, @Addr, @AddrLen);

    if ClientSocket = INVALID_SOCKET then //        FSocket
    begin
      if not T.Stopped then StopListen;
      Break;
    end;
    
    //       ,    -
    //   (, ,    )
    if T.Stopped then
    begin
      shutdown(ClientSocket, SD_BOTH);
      closesocket(ClientSocket);
      Break;
    end
    else
    begin
      FRemAddr := Integer(Addr.sin_addr);
      FRemPort := ntohs(Addr.sin_port);
      if AsyncEvents then
        _SyncClientConnect(Pointer(ClientSocket))
      else
        T.Synchronize(_SyncClientConnect, Pointer(ClientSocket));
    end;
  end;
end;

procedure TBasicTCPServer._SyncClientConnect(Param: Pointer); 
begin
  ConnectionAccepted(TSocket(Param), FRemAddr, FRemPort);
end;

function TBasicTCPServer.GetIsListening: Boolean;
begin
  Result := FSocket <> 0;
end;

procedure TBasicTCPServer.ConnectionAccepted(ClientSocket: TSocket; RemAddr: Integer; RemPort: Integer);
begin
  shutdown(ClientSocket, SD_BOTH);
  closesocket(ClientSocket);
end;

function TBasicTCPServer.StartServer(Port: Integer): Integer;
var 
  Addr: TSockAddrIn;
label
  ErrorLbl;
begin
  Result := -1;
  if FSocket <> 0 then Exit;

  FSocket := socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
  if FSocket = 0 then goto ErrorLbl;

  
  Addr.sin_family := AF_INET;
  Addr.sin_port := htons(Port);
  Addr.sin_addr.S_addr := INADDR_ANY;
  
  if bind(FSocket, Addr, SizeOf(Addr)) = SOCKET_ERROR then goto ErrorLbl;

  
  if listen(FSocket, SOMAXCONN) = SOCKET_ERROR then goto ErrorLbl;

  FAcceptThread := TThread.Create;
  FAcceptThread.OnExecute := _ExecuteAccept;
  FAcceptThread.AutoFree := True;
  FAcceptThread.Resume;

  
  Result := 0;
  Exit;
  
ErrorLbl:
  Result := WSAGetLastError;
  closesocket(FSocket);
  FSocket := 0;
end;

procedure TBasicTCPServer.StopListen;
begin
  if FAcceptThread <> nil then
  begin
    FAcceptThread.Stop;
    FAcceptThread := nil;
  end;
  
  if FSocket <> 0 then
  begin
    shutdown(FSocket, SD_BOTH);
    closesocket(FSocket);
    FSocket := 0;
  end;
end;


// ====== TCustomClientConnection ====== //

constructor TCustomClientConnection.Create(Socket: TSocket; RemIP: Integer; RemPort: Integer = 0);
begin
  inherited Create;
  FSocket := Socket;
  FRemotePort := RemPort;
  FRemoteAddress := IPv4AddressToString(RemIP);
end;

procedure TCustomClientConnection.AfterClose(Reason: Integer);
begin
  if Server <> nil then Server.ConnectionDelete(Self);
end;
  
  
// ====== TCustomTCPServer ====== //

constructor TCustomTCPServer.Create;
begin
  inherited;
  InitializeCriticalSection(FCSList);
  FConnList := NewList;
  FConnList.AddBy := 20;
end;

destructor TCustomTCPServer.Destroy;
begin
  StopServer;

  EnterCriticalSection(FCSList);
    FConnList.Free;
  LeaveCriticalSection(FCSList);
  
  DeleteCriticalSection(FCSList);
  inherited;
end;

procedure TCustomTCPServer.AfterNewConnection(Conn: TCustomClientConnection);
begin

end;

function TCustomTCPServer.BeforeNewConnection(ClientSocket: TSocket; RemAddr: Integer; RemPort: Integer): TCustomClientConnection;
begin
  Result := nil;
end;

procedure TCustomTCPServer.ConnectionAccepted(ClientSocket: TSocket; RemAddr: Integer; RemPort: Integer);
var
  Connection: TCustomClientConnection;
begin
  Connection := BeforeNewConnection(ClientSocket, RemAddr, RemPort);
  if Connection = nil then
  begin
    inherited;
    Exit;
  end;
  ConnectionAdd(Connection);
  AfterNewConnection(Connection);
end;


procedure TCustomTCPServer.ConnectionAdd(Conn: TCustomClientConnection);
begin
  EnterCriticalSection(FCSList);
    if FConnList.IndexOf(Conn) = -1 then
    begin
      Conn.Server := Self;
      FConnList.Add(Conn);
    end;
  LeaveCriticalSection(FCSList);
end;

procedure TCustomTCPServer.ConnectionDelete(Conn: TCustomClientConnection);
begin
  EnterCriticalSection(FCSList);
    FConnList.Remove(Conn);
    //Conn.DecRef; //    
  LeaveCriticalSection(FCSList);
end;

function TCustomTCPServer.GetCount: Integer;
begin
  Result := FConnList.Count;
end;

procedure TCustomTCPServer.CloseAll;
var 
  C: TCustomClientConnection;
  I: Integer;
begin
  EnterCriticalSection(FCSList);
    for I := FConnList.Count - 1 downto 0 do
    begin
      C := FConnList.Items[I];
      C.Disconnect; // TCustomClientConnection.AfterClose     ,  
      //C.DecRef;
    end;
    FConnList.Clear;
  LeaveCriticalSection(FCSList);
end;

function TCustomTCPServer.ConnectionExists(Conn: TCustomClientConnection): Boolean;
begin
  EnterCriticalSection(FCSList);
    Result := (FConnList.IndexOf(Conn) > -1);
  LeaveCriticalSection(FCSList);
end;

procedure TCustomTCPServer.GetConnections(OutConn: PList = nil; OutIP: PStrList = nil);
var 
  C: TCustomClientConnection;
  I: Integer;
begin
  EnterCriticalSection(FCSList);
    for I := 0 to FConnList.Count - 1 do
    begin
      C := FConnList.Items[I];
      if Assigned(OutConn) then OutConn.Add(C);
      if Assigned(OutIP) then OutIP.Add(C.RemoteAddress);
    end;
  LeaveCriticalSection(FCSList);
end;

function TCustomTCPServer.HoldConnection(Conn: TCustomClientConnection): TCustomClientConnection;
begin
  EnterCriticalSection(FCSList);
    if FConnList.IndexOf(Conn) > -1 then
    begin
      Result := Conn;
      Conn.IncRef;
    end
    else
      Result := nil;
  LeaveCriticalSection(FCSList);
end;

procedure TCustomTCPServer.StopSendAll;
var 
  I: Integer;
begin
  EnterCriticalSection(FCSList);  
    //TODO:
    //   SendManager      .
    for I := 0 to FConnList.Count - 1 do
    begin
      SendManager.DeleteConnection(TCustomConnection(FConnList.Items[I]));
    end;
  LeaveCriticalSection(FCSList);
end;

procedure TCustomTCPServer.StopServer;
begin
  StopListen;
  CloseAll;
end;


// ====== TSharedString ====== //

constructor TSharedString.Create(const S: string);
begin
  inherited Create;
  FString := S;
end;

function TSharedString.GetBuffer(Offset: Integer): Pointer;
begin
  Result := Pointer(Integer(Pointer(FString)) + Offset);
end;

function TSharedString.GetLength: Integer;
begin
  Result := Length(FString);
end;


// ====== TSendDataString ====== //

constructor TSendDataString.Create(Data: TSharedString);
begin
  inherited Create;
  FData := Data;
  FLength := Data.GetLength;
  FSended := 0;
end;

destructor TSendDataString.Destroy;
begin
  Finish;
  inherited;
end;

procedure TSendDataString.Finish;
begin
  if FData <> nil then
  begin
    FData.DecRef;
    FData := nil;
    //FLength := 0;
  end;
end;

function TSendDataString.GetNextData(var Buffer: Pointer; NeedLen: Integer): Integer;
begin
  Result := 0;

  if Stopped or (FData = nil) then Exit;
  
  Result := Min(NeedLen, FLength);
  if Result = 0 then Exit;
  
  Buffer := FData.GetBuffer(FSended);
  Inc(FSended, Result);
  Dec(FLength, Result);
end;

// ====== TTCPClient ====== //

procedure TTCPClient.AfterClose(Reason: Integer);
begin
  if Assigned(OnDisconnect) then OnDisconnect(Self);
end;

procedure TTCPClient.DataReceived(Data: TReceiveBuffer);
begin
  if Assigned(OnReceive) then OnReceive(Self, Data.GetString);
end;

procedure TTCPClient.SendResult(Data: TSendData; ResultCode: Integer);
begin
  if ResultCode = 0 then
  begin
    if Assigned(OnSend) then OnSend(Self, Data);
  end
  else
  begin
    if Assigned(OnErrorSend) then OnErrorSend(Self, ResultCode);
  end;
end;


// ====== TClientConnection ====== //

procedure TClientConnection.AfterClose(Reason: Integer);
begin
  inherited;
  if Assigned(TTCPServer(Server).OnClientDisconnect) then 
    TTCPServer(Server).OnClientDisconnect(Self);
  DecRef; //  
end;

procedure TClientConnection.DataReceived(Data: TReceiveBuffer);
begin
  if Assigned(TTCPServer(Server).OnReceive) then TTCPServer(Server).OnReceive(Self, Data.GetString);
end;

procedure TClientConnection.SendResult(Data: TSendData; ResultCode: Integer);
begin
  if (ResultCode = 0) then
  begin
    if Assigned(TTCPServer(Server).OnSend) then TTCPServer(Server).OnSend(Self, Data);
  end
  else
  begin
    if Assigned(TTCPServer(Server).OnErrorSend) then TTCPServer(Server).OnErrorSend(Self, ResultCode);
  end;
end;

// ====== TTCPServer ====== //

function TTCPServer.BeforeNewConnection(ClientSocket: TSocket; RemAddr: Integer; RemPort: Integer): TCustomClientConnection;
begin
  Result := TClientConnection.Create(ClientSocket, RemAddr, RemPort);
  Result.AsyncEvents := Self.AsyncEvents;
  Result.OnProgress := Self.OnProgress;
end;

procedure TTCPServer.AfterNewConnection(Conn: TCustomClientConnection);
begin
  if Assigned(OnClientConnect) then OnClientConnect(Conn);
end;

procedure TTCPServer.SendAll(Data: TSharedString);
var
  Lst: PList;
  I: Integer;
  C: TCustomClientConnection;
begin
  Lst := NewList;
  GetConnections(Lst); //    
    for I := 0 to Lst.Count - 1 do
    begin
      C := HoldConnection(TCustomClientConnection(Lst.Items[I]));
      if C <> nil then
      begin
        Data.IncRef;
        C.Send(TSendDataString.Create(Data));
        C.DecRef;
      end;
    end;
    Data.DecRef;
  Lst.Free;
end;

procedure TTCPServer.SendAllAsync(Data: TSharedString; Queue: Boolean = True);
var
  Lst: PList;
  I: Integer;
  C: TCustomClientConnection;
begin
  Lst := NewList;
  GetConnections(Lst); //    
    for I := 0 to Lst.Count - 1 do
    begin
      C := HoldConnection(TCustomClientConnection(Lst.Items[I]));
      if C <> nil then
      begin
        if Queue or not C.IsSending then
        begin
          Data.IncRef;
          SendManager.SendAsync(C, TSendDataString.Create(Data));
        end;
        C.DecRef;
      end;
    end;
    Data.DecRef;
  Lst.Free;
end;











initialization
  WSAStartup($101, wsaData);
  
  ReceivingManager := TReceivingManager.Create;
  SendManager := TSendManager.Create;

finalization
  ReceivingManager.Free;
  SendManager.Free;

end.