// ======================================= //
//                                         //
//             //
//            2017-03-27             //
//                                         //
//               HiAsm               //
//             www.hiasm.com               //
//                                         //
//            : Netspirit             //
//        E-mail: netspirit@meta.ua        //
//                                         //
// ======================================= //


unit NSThreads;

interface


uses
  Windows;

type

  //     -     
  //   Free/Destroy  DecRef
  TSharedObject = class
    private
      FRefCount: Integer;
      function GetRefCount: Integer;
      procedure SetRefCount(Value: Integer);
    public
      property RefCount: Integer read GetRefCount write SetRefCount;
      
      function IncRef: Integer; //       -  = 1.   
      function DecRef: Integer; //    0    .   
  end;


  TCustomThread = class;
  
  TSyncMethod = procedure (Param: Pointer) of object;
  TThreadMethod = procedure (Sender: TCustomThread; Param: Pointer) of object;
  
  TSyncRecord = record
    Thread: TCustomThread;
    Param: Pointer;
    case Integer of
      0: (ThreadMethod: TThreadMethod);
      1: (SyncMethod: TSyncMethod);
  end;
  PSyncRecord = ^TSyncRecord;
  
  
  //    
  //    DecRef (!)
  //        Execute()
  TCustomThread = class(TSharedObject)
    private
      
    protected
      FThreadHandle: THandle;
      FShouldStop: Boolean;
      FSuspended: Boolean;
      FThreadId: DWORD;
      FTerminated: Boolean;
      
      procedure Execute; virtual; //  ,    
      procedure InternalExecute;
      procedure PrepareThread;
      
    public
      property Stopped: Boolean read FShouldStop;
      property Suspended: Boolean read FSuspended;
      property Terminated: Boolean read FTerminated;
      property ThreadHandle: THandle read FThreadHandle;
      property ThreadId: DWORD read FThreadId;
      
      constructor Create;
      destructor Destroy; override;
      
      procedure Resume; //      Suspend
      procedure Restart; //     ,    Terminated (    TThread)
      procedure Stop; //  Stopped=True. OnExecute         
      procedure StopAndWait; //  Stop,    
      procedure Suspend; //   ( )
      procedure Synchronize(Method: TSyncMethod; Param: Pointer = nil);
      procedure SynchronizeEx(Method: TThreadMethod; Param: Pointer = nil);
      procedure Terminate; //   .        (!)
  
  end;
  
  
  TOnThreadExecute = function (Sender: TCustomThread): Integer of object;
  
  //  KOL.TThread
  TThread = class(TCustomThread)
  
    private
      
    protected
      procedure Execute; override;
      
    public
      AutoFree: Boolean;
      FResult: Integer;
      Data: Pointer;
      Tag: Integer;

      OnExecute: TOnThreadExecute;
  end;
  
  

  
  procedure CallSynchronized(Method: TSyncMethod; Param: Pointer = nil); //      .     TThread
  
  // ,   TThread.Synchronize     
  // TODO:    
  function SyncAvailable: Boolean;
  
var
  //   (    Applet), 
  //       
  ThreadSyncRoutine: procedure (const SyncRec: TSyncRecord);
  
  
  
  
implementation

uses 
  Kol, Messages;
  
const
  AM_THREAD_SYNC = WM_APP + 1; // ,       


var
  ProcAttached: Boolean;
  

function WndProcThreadSync(Sender: PControl; var Msg: TMsg; var Rslt: Integer): Boolean;
var 
  SR: TSyncRecord;
begin
  Result := False;
  if Msg.message = AM_THREAD_SYNC then
  begin
    SR := PSyncRecord(Msg.lParam)^;
    if Assigned(SR.Thread) then //   SR.Thread,    SR -  TThreadMethod
    begin
      SR.ThreadMethod(SR.Thread, SR.Param);
    end
    else
    begin
      SR.SyncMethod(SR.Param);
    end;
    Rslt := 0;
  end;
end;

procedure CallSynchronized(Method: TSyncMethod; Param: Pointer = nil);
var
  SR: TSyncRecord;
begin
  SR.Thread := nil;
  SR.Param := Param;
  SR.SyncMethod := Method;
  
  if Assigned(Applet) then
  begin
    if not ProcAttached then
    begin
      Applet.AttachProc(WndProcThreadSync);
      ProcAttached := True;
    end;
    
    SendMessage(Applet.Handle, AM_THREAD_SYNC, 0, Integer(@SR));
  end
  else
  begin
    if Assigned(ThreadSyncRoutine) then
      ThreadSyncRoutine(SR)
    else //     -   
      Method(Param);
  end;
end;

function SyncAvailable: Boolean;
begin
  Result := Assigned(Applet) or Assigned(ThreadSyncRoutine);
end;


// ====== TSharedObject ====== //

function TSharedObject.GetRefCount: Integer;
begin
  Result := FRefCount + 1;
end;

procedure TSharedObject.SetRefCount(Value: Integer);
begin
  FRefCount := Value - 1;
end;

function TSharedObject.IncRef: Integer;
begin
  Result := InterlockedIncrement(FRefCount) + 1;
end;

// ,      FRefCount = -1,
//  FRefCount = 0   , 1 -   ..
//  ,   inherited Create     .
//       ""    1  FRefCount,
//    " ".

function TSharedObject.DecRef: Integer;
begin
  Result := InterlockedDecrement(FRefCount) + 1;
  if Result < 1 then Destroy; //Free;

  // :   Result = 0,
  //    DecRef    .
  //        DecRef 
  //  ,   RefCount,   
  //       .     .
end;


// ====== TCustomThread ====== //

function ThreadFunc(Thread: TCustomThread): Integer; {$ifdef F_P}stdcall;{$endif}
begin
  Result := 0;
  Thread.InternalExecute;
end;

constructor TCustomThread.Create;
begin
  inherited;
  
  if (not ProcAttached) and (Applet <> nil) then
  begin
    Applet.AttachProc(WndProcThreadSync);
    ProcAttached := True;
  end;
  
  PrepareThread;
end;

destructor TCustomThread.Destroy;
begin
  if not FTerminated then
  begin
    if not FSuspended then
    begin
      Stop;
      WaitForSingleObject(FThreadHandle, 200); //    (?)
    end;

    Terminate;
    WaitForSingleObject(FThreadHandle, INFINITE); // ?
  end;
  CloseHandle(FThreadHandle);
  
  inherited;
end;

procedure TCustomThread.Execute;
begin

end;

procedure TCustomThread.InternalExecute;
begin
  IncRef; //  ,     ,   Execute()
  
  Execute; //      
  
  FTerminated := True; //    Terminate  Destroy
  
  DecRef; //  
end;

procedure TCustomThread.PrepareThread;
begin
  if FThreadHandle <> 0 then CloseHandle(FThreadHandle);
  
  FThreadHandle := {$ifdef F_P}CreateThread{$else}BeginThread{$endif}
    (
      nil, // no security
      0,   // the same stack size
      @ThreadFunc, // thread entry point
      Self,      // parameter to pass to ThreadFunc
      CREATE_SUSPENDED,   // always SUSPENDED
      Self.FThreadID // receive thread ID
    );
  
  if FThreadHandle <> 0 then
  begin
    FSuspended := True;
    FTerminated := False;
  end
  else
  begin
    FSuspended := False;
    FTerminated := True;      
  end;
end;

procedure TCustomThread.Stop;
begin
  FShouldStop := True;
end;

procedure TCustomThread.StopAndWait;
begin
  FShouldStop := True;
  WaitForSingleObject(FThreadHandle, INFINITE);
end;

procedure TCustomThread.Resume;
begin
  FSuspended := False;
  if FTerminated then Exit;
  
  if (ResumeThread(FThreadHandle) > 1) then FSuspended := True;
end;

procedure TCustomThread.Restart;
begin
  FShouldStop := False;
  if FTerminated then PrepareThread;
  if FSuspended then Resume;
end;

procedure TCustomThread.Suspend;
begin
  if (not FSuspended) and (not FTerminated) then
  begin
    FSuspended := True;
    SuspendThread(FThreadHandle);
  end;
end;

procedure TCustomThread.Synchronize(Method: TSyncMethod; Param: Pointer = nil);
var
  SR: TSyncRecord;
begin
  SR.Thread := nil;
  SR.Param := Param;
  SR.SyncMethod := Method;
  
  if Assigned(Applet) then
    SendMessage(Applet.Handle, AM_THREAD_SYNC, 0, Integer(@SR))
  else
  begin
    if Assigned(ThreadSyncRoutine) then
      ThreadSyncRoutine(SR)
    else //     -   
      Method(Param);
  end;
end;

procedure TCustomThread.SynchronizeEx(Method: TThreadMethod; Param: Pointer = nil);
var
  SR: TSyncRecord;
begin
  SR.Thread := Self;
  SR.Param := Param;
  SR.ThreadMethod := Method;
  
  if Assigned(Applet) then
    SendMessage(Applet.Handle, AM_THREAD_SYNC, 0, Integer(@SR))
  else
  begin
    if Assigned(ThreadSyncRoutine) then
      ThreadSyncRoutine(SR)
    else //     -   
      Method(Self, Param);
  end;
end;

procedure TCustomThread.Terminate;
begin
  if not FTerminated then
  begin
    FTerminated := True;
    TerminateThread(FThreadHandle, 0);
    if not FSuspended then //   ( IncRef()  InternalExecute())
      DecRef //    ,  DecRef()  InternalExecute()    (   ...)
    else
      FSuspended := False;
  end;
end;

// ====== TThread ====== //

procedure TThread.Execute;
begin
  FResult := 0;
  
  if Assigned(OnExecute) then FResult := OnExecute(Self);
  
  if AutoFree then DecRef; //    AutoFree
end;


end.