unit hiTCPClient;

interface


uses 
  Windows, Kol, Share, NSTCP, NSThreads;


type

  ThiTCPClient = class(TTCPClient)
    private
      FConnectThread: TThread;
      FPort: Integer;
      FIP: string;
      
      F_event_onDisconnect: THI_Event;
      F_event_onErrorSend: THI_Event;
      F_event_onSend: THI_Event;
      F_event_onProgress: THI_Event;
      F_event_onReceive: THI_Event;
      
      function _ExecuteConnect(T: TCustomThread): Integer;
      procedure _OnDisconnect(Sender: TCustomConnection);
      procedure _OnProgress(Sender: TCustomConnection; Data: TSendData);
      procedure _OnReceive(Sender: TCustomConnection; Data: string);
      procedure _OnErrorSend(Sender: TCustomConnection; ErrorCode: Integer);
      procedure _OnSend(Sender: TCustomConnection; Data: TSendData);
      procedure _SyncConnectResult(Param: Pointer);
      
      procedure AssignEvent(Index: Integer; Value: THI_Event);
      procedure SetAsyncEvents(Value: Integer);

    public
      _prop_IP: string;
      _prop_Port: Integer;
      _prop_AsyncConnect: Integer;
      _prop_AutoReceive: Integer;
      _prop_OverSend: Integer;
      _prop_SendTimeout: Integer;
      
      _data_IP: THI_Event;
      _data_Port: THI_Event;
      _data_Data: THI_Event;
      
      _event_onConnect: THI_Event;
      _event_onErrorConnect: THI_Event;

      property _prop_AsyncEvents: Integer write SetAsyncEvents;

      
      property _event_onDisconnect: THI_Event index 1 write AssignEvent;
      property _event_onErrorSend: THI_Event index 2 write AssignEvent;
      property _event_onProgress: THI_Event index 3 write AssignEvent;
      property _event_onSend: THI_Event index 4 write AssignEvent;
      property _event_onReceive: THI_Event index 5 write AssignEvent;
      
      destructor Destroy; override;
      
      procedure _work_doAbortSend(var _Data: TData; Index: Word);
      procedure _work_doConnect(var _Data: TData; Index: Word);
      procedure _work_doDisconnect(var _Data: TData; Index: Word);
      procedure _work_doSend(var _Data: TData; Index: Word);
      procedure _work_doSendAsync(var _Data: TData; Index: Word);
      procedure _work_doStartReceiving(var _Data: TData; Index: Word);
      procedure _work_doStopReceiving(var _Data: TData; Index: Word);
  end;
  
  //    TSharedString  TCPClientServer.TSendDataString -   
  // TODO:  TCPClientServer.TSendDataString
  TSendDataString = class(TSendData)
    private
      FString: string;
    public
      constructor Create(const S: string);
      procedure Finish; override;
      function GetNextData(var Buffer: Pointer; NeedLen: Integer): Integer; override;
  end;

  
implementation


constructor TSendDataString.Create(const S: string);
begin
  inherited Create;
  FString := S;
  FLength := Length(FString);
  FSended := 0;
end;

procedure TSendDataString.Finish;
begin
  FString := '';
  FLength := 0;
end;

function TSendDataString.GetNextData(var Buffer: Pointer; NeedLen: Integer): Integer;
begin
  Result := 0;
  if Stopped then Exit;
  
  Result := Min(NeedLen, FLength);
  if Result = 0 then Exit;
  
  Buffer := @FString[1+FSended];
  Inc(FSended, Result);
  Dec(FLength, Result);
end;

destructor ThiTCPClient.Destroy;
begin
  if FConnectThread <> nil then
  begin
    FConnectThread.Free;
    FConnectThread := nil;
  end;
  OnDisconnect := nil; //  Disconnect    -   
  Disconnect;
  inherited;
end;

function ThiTCPClient._ExecuteConnect(T: TCustomThread): Integer;
begin
  Result := Connect(FIP, FPort);

  if (T <> nil) then
  begin
    if not T.Stopped then 
    begin
      FConnectThread := nil;
      if AsyncEvents then
        _SyncConnectResult(Pointer(Result))
      else
        T.Synchronize(_SyncConnectResult, Pointer(Result));
    end;
  end
  else
    _SyncConnectResult(Pointer(Result));
end;

procedure ThiTCPClient._OnDisconnect(Sender: TCustomConnection);
begin
  _hi_OnEvent(F_event_onDisconnect);
end;

procedure ThiTCPClient._OnProgress(Sender: TCustomConnection; Data: TSendData);
begin
  _hi_OnEvent(F_event_onProgress, Data.Sended);
end;

procedure ThiTCPClient._OnReceive(Sender: TCustomConnection; Data: string);
begin
  _hi_OnEvent(F_event_onReceive, Data);
end;

procedure ThiTCPClient._OnSend(Sender: TCustomConnection; Data: TSendData);
begin
  _hi_OnEvent(F_event_onSend);
end;

procedure ThiTCPClient._OnErrorSend(Sender: TCustomConnection; ErrorCode: Integer);
begin
  _hi_OnEvent(F_event_onErrorSend, ErrorCode);
end;

procedure ThiTCPClient._SyncConnectResult(Param: Pointer);
begin
  if Integer(Param) = 0 then
  begin
    SendTimeout := _prop_SendTimeout;
    _hi_OnEvent(_event_onConnect);
    if _prop_AutoReceive <> 0 then StartReceiving;
  end
  else
    _hi_OnEvent(_event_onErrorConnect, Integer(Param));
end;

procedure ThiTCPClient.AssignEvent(Index: Integer; Value: THI_Event);
begin
  case Index of
    1:
      begin
        F_event_onDisconnect := Value;
        OnDisconnect := _OnDisconnect;
      end;
    2:
      begin
        F_event_onErrorSend := Value;
        OnErrorSend := _OnErrorSend;
      end;
    3:
      begin
        F_event_onProgress := Value;
        OnProgress := _OnProgress;
      end;
    4:
      begin
        F_event_onSend := Value;
        OnSend := _OnSend;
      end;
    5:
      begin
        F_event_onReceive := Value;
        OnReceive := _OnReceive;
      end;
  end;
end;

procedure ThiTCPClient.SetAsyncEvents(Value: Integer);
begin
  AsyncEvents := Boolean(Value);
end;



procedure ThiTCPClient._work_doAbortSend;
begin
  StopSend;
end;

procedure ThiTCPClient._work_doConnect;
begin
  if (FConnectThread <> nil) or Connected then Exit; //  ,   
  
  FIP := ReadString(_Data, _data_IP, _prop_IP);
  FPort := ReadInteger(_Data, _data_Port, _prop_Port);
  
  if _prop_AsyncConnect <> 0 then
  begin
    FConnectThread := TThread.Create;
    FConnectThread.OnExecute := _ExecuteConnect;
    FConnectThread.AutoFree := True;
    FConnectThread.Resume;
  end
  else
    _ExecuteConnect(nil);
end;

procedure ThiTCPClient._work_doDisconnect;
begin
  if FConnectThread <> nil then
  begin
    FConnectThread.Free;
    FConnectThread := nil;
  end;
  Disconnect;
end;


procedure ThiTCPClient._work_doSend;
begin
  Send(TSendDataString.Create(ReadString(_Data, _data_Data, '')));
end;

procedure ThiTCPClient._work_doSendAsync;
begin
  if (_prop_OverSend = 0) and IsSending then Exit;
  SendAsync(TSendDataString.Create(ReadString(_Data, _data_Data, '')));
end;

procedure ThiTCPClient._work_doStartReceiving;
begin
  StartReceiving;
end;

procedure ThiTCPClient._work_doStopReceiving;
begin
  StopReceiving;
end;


end.