unit hiTCPServer;

interface


uses 
  Windows, Kol, Share, winsock, NSTCP, NSThreads;



type
  
  ThiTCPServer = class(TTCPServer)
    private
      FCSClient: TRTLCriticalSection;
      
      FCurClientID: Integer;
      FCurIPAddress: string;
      NoSync: Boolean;

      procedure _OnClientConnect(Conn: TCustomClientConnection);
      procedure _OnClientDisconnect(Conn: TCustomClientConnection);
      procedure _OnErrorSend(Conn: TCustomClientConnection; ErrorCode: Integer);
      procedure _OnProgress(Conn: TCustomConnection; Data: TSendData);
      procedure _OnReceive(Conn: TCustomClientConnection; S: string);
      procedure _OnSend(Conn: TCustomClientConnection; Data: TSendData);
      
      procedure SetAsyncEvents(Value: Integer);

    public
      _prop_Port: Integer;
      _prop_OverSend: Integer;
      _prop_SendTimeout: Integer;
      
      _data_Data: THI_Event;
      _data_ClientID: THI_Event;
      _data_Port: THI_Event;
            
      _event_onEnumClients: THI_Event;
      _event_onErrorStart: THI_Event;
      _event_onStartServer:THI_Event;
      
      _event_onClientConnect: THI_Event;
      _event_onClientDisconnect: THI_Event;
      _event_onErrorSend: THI_Event;
      _event_onProgress: THI_Event;
      _event_onReceive: THI_Event;
      _event_onSend: THI_Event;
      
      property _prop_AsyncEvents: Integer write SetAsyncEvents;

      constructor Create;
      destructor Destroy; override;

      procedure _work_doAbortSend(var _Data: TData; Index: Word);
      procedure _work_doAbortSendAll(var _Data: TData; Index: Word);
      procedure _work_doClose(var _Data: TData; Index: Word);
      procedure _work_doCloseAll(var _Data: TData; Index: Word);
      procedure _work_doEnumClients(var _Data: TData; Index: Word);
      procedure _work_doSend(var _Data: TData; Index: Word);
      procedure _work_doSendAll(var _Data: TData; Index: Word);
      procedure _work_doSendAllAsync(var _Data: TData; Index: Word);
      procedure _work_doSendAsync(var _Data: TData; Index: Word);
      procedure _work_doStartServer(var _Data: TData; Index: Word);
      procedure _work_doStopListen(var _Data: TData; Index: Word);
      procedure _work_doStopServer(var _Data: TData; Index: Word);

      procedure _var_Count(var _Data: TData; Index: Word);
      procedure _var_CurClientID(var _Data: TData; Index: Word);
      procedure _var_CurIPAddress(var _Data: TData; Index: Word);
  end;


implementation


constructor ThiTCPServer.Create;
begin
  inherited;
  InitializeCriticalSection(FCSClient);
  OnClientConnect := _OnClientConnect;
  OnClientDisconnect := _OnClientDisconnect;
end;

destructor ThiTCPServer.Destroy;
begin
  if NoSync then EnterCriticalSection(FCSClient);
    OnClientDisconnect := nil;
    StopServer;
  if NoSync then LeaveCriticalSection(FCSClient);
  DeleteCriticalSection(FCSClient);
  inherited;
end;

procedure ThiTCPServer._OnClientConnect(Conn: TCustomClientConnection);
begin
  if NoSync then EnterCriticalSection(FCSClient);
    Conn.SendTimeout := _prop_SendTimeout;
    
    FCurClientID := Integer(Conn);
    FCurIPAddress := Conn.RemoteAddress;
    
    _hi_OnEvent(_event_onClientConnect, Integer(Conn));
  if NoSync then LeaveCriticalSection(FCSClient);
  
  if ConnectionExists(Conn) then Conn.StartReceiving;
end;

procedure ThiTCPServer._OnClientDisconnect(Conn: TCustomClientConnection);
begin
  if NoSync then EnterCriticalSection(FCSClient);
    FCurClientID := Integer(Conn);
    _hi_OnEvent(_event_onClientDisconnect, Integer(Conn));
  if NoSync then LeaveCriticalSection(FCSClient);
end;

procedure ThiTCPServer._OnErrorSend(Conn: TCustomClientConnection; ErrorCode: Integer);
begin
  if NoSync then EnterCriticalSection(FCSClient);
    FCurClientID := Integer(Conn);
    _hi_OnEvent(_event_onErrorSend, ErrorCode);
  if NoSync then LeaveCriticalSection(FCSClient);
end;

procedure ThiTCPServer._OnProgress(Conn: TCustomConnection; Data: TSendData);
begin
  if NoSync then EnterCriticalSection(FCSClient);
    FCurClientID := Integer(Conn);
    _hi_OnEvent(_event_onProgress, Data.Sended);
  if NoSync then LeaveCriticalSection(FCSClient);
end;

procedure ThiTCPServer._OnReceive(Conn: TCustomClientConnection; S: string);
begin
  if NoSync then EnterCriticalSection(FCSClient);
    FCurClientID := Integer(Conn);
    _hi_OnEvent(_event_onReceive, S);
  if NoSync then LeaveCriticalSection(FCSClient);
end;

procedure ThiTCPServer._OnSend(Conn: TCustomClientConnection; Data: TSendData);
begin
  if NoSync then EnterCriticalSection(FCSClient);
    FCurClientID := Integer(Conn);
    _hi_OnEvent(_event_onSend, Integer(Conn));
  if NoSync then LeaveCriticalSection(FCSClient);
end;

procedure ThiTCPServer.SetAsyncEvents(Value: Integer);
begin
  AsyncEvents := Boolean(Value);
  NoSync := (not SyncAvailable) or AsyncEvents;
end;


procedure ThiTCPServer._work_doAbortSend;
var
  C: TCustomClientConnection;
begin
  C := HoldConnection(TCustomClientConnection(ReadInteger(_Data, _data_ClientID, 0)));
  if C <> nil then 
  begin
    C.StopSend;
    C.DecRef;
  end;
end;

procedure ThiTCPServer._work_doAbortSendAll;
begin
  StopSendAll;
end;

procedure ThiTCPServer._work_doClose;
var
  C: TClientConnection;
begin
  C := TClientConnection(HoldConnection(TClientConnection(ReadInteger(_Data, _data_ClientID, 0))));
  if C <> nil then 
  begin
    C.Disconnect;
    C.DecRef;
  end;
end;

procedure ThiTCPServer._work_doCloseAll;
begin
  CloseAll;
end;

procedure ThiTCPServer._work_doEnumClients;
var
  Lst: PList;
  IPLst: PStrList;
  i: Integer;
  Conn: TClientConnection;
begin
  Lst := NewList;
  IPLst := NewStrList;
  GetConnections(Lst, IPLst);
    for i := 0 to Lst.Count-1 do
    begin
      Conn := TClientConnection(Lst.Items[i]);
      if NoSync then EnterCriticalSection(FCSClient);
        FCurClientID := Integer(Conn);
        FCurIPAddress := IPLst.Items[i];
        _hi_OnEvent(_event_onEnumClients, FCurClientID);
      if NoSync then LeaveCriticalSection(FCSClient);
    end;
  Lst.Free;
  IPLst.Free;
end;

procedure ThiTCPServer._work_doSend;
var
  D: TSharedString;
  C: TClientConnection;
begin
  D := TSharedString.Create(ReadString(_Data, _data_Data, ''));
  C := TClientConnection(HoldConnection(TClientConnection(ReadInteger(_Data, _data_ClientID, 0))));
  if C <> nil then 
  begin
    C.Send(TSendDataString.Create(D));
    C.DecRef;
  end
  else
    D.DecRef;
end;

procedure ThiTCPServer._work_doSendAll;
begin
  SendAll(TSharedString.Create(ReadString(_Data, _data_Data, '')));
end;

procedure ThiTCPServer._work_doSendAllAsync;
begin
  SendAllAsync(TSharedString.Create(ReadString(_Data, _data_Data, '')), (_prop_OverSend = 1));
end;

procedure ThiTCPServer._work_doSendAsync;
var
  D: TSharedString;
  C: TCustomConnection;
begin
  D := TSharedString.Create(ReadString(_Data, _data_Data, ''));
  C := HoldConnection(TClientConnection(ReadInteger(_Data, _data_ClientID, 0)));
  
  if (C = nil) or ((_prop_OverSend = 0) and C.IsSending) then 
  begin
    D.DecRef;
  end
  else
  begin
    C.SendAsync(TSendDataString.Create(D));
    C.DecRef;
  end;
end;

procedure ThiTCPServer._work_doStartServer;
var
  Rslt: Integer;
begin
  if IsListening then Exit;
  
  if Assigned(_event_onErrorSend.Event) then OnErrorSend := _OnErrorSend;
  if Assigned(_event_onProgress.Event) then OnProgress := _OnProgress;
  if Assigned(_event_onReceive.Event) then OnReceive := _OnReceive;
  if Assigned(_event_onSend.Event) then OnSend := _OnSend;
  
  Rslt := StartServer(ReadInteger(_Data, _data_Port, _prop_Port));
  if Rslt = 0 then
    _hi_CreateEvent(_Data, @_event_onStartServer)
  else
    _hi_CreateEvent(_Data, @_event_onErrorStart, Rslt);
end;

procedure ThiTCPServer._work_doStopListen;
begin
  StopListen;
end;

procedure ThiTCPServer._work_doStopServer;
begin
  StopServer;
end;



procedure ThiTCPServer._var_Count;
begin
  dtInteger(_Data, Count);
end;

procedure ThiTCPServer._var_CurClientID;
begin
  dtInteger(_Data, FCurClientID);
end;

procedure ThiTCPServer._var_CurIPAddress;
begin
  dtString(_Data, FCurIPAddress);
end;


end.